/*
 * main.c
 *
 *  original version:
 *  Christophe Lucas <c.lucas@ifrance.com>
 *  new release:
 *  Luca Amore (lookee) <luca@lookee.it>
 *  This program is relased under GNU/GPL term :-)
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>

#include <linux/compiler.h>
#include <linux/videodev.h>

#include "config.h"
#include "capture.h"
#include "ftpup.h"


#define CONF_FILE "webcam.conf"
#define PLUS	1
#define MINUS	-1


#define STR_DIV "--------------------------------------------------\n" 

int ftp_upload=1; 	/* (1 enabled|0 disabled) ftp upload */
int ftp_monitor=0;	/* (1 enabled|0 disabled) ftp upload monitor */
int save_img=1;		/* (1 enabked!0 disabled) image saving in current directory */

/* internal */
static unsigned int ftp_rate_counter=1;

/*
 * FIX ME:
 * Use change_ctrl as soon as possible
 */
#define STEP	4

#include <SDL/SDL.h>

/* configuration keys readed from file */

struct config_t *config;

extern void bayer2rgb24(unsigned char *dst, unsigned char *src, long int WIDTH, long int HEIGHT) ;
extern void enumerate_controls_on_device( int fd ) ;
extern int change_ctrl( int fd , __u32 ctrl_id, int down_up ) ;


int main( int argc, char **argv)
{
    int			fd1, done ;
    unsigned long int	x, y, f;
    unsigned char	*s, *d;

    Uint32		rmask, gmask, bmask, amask;
    Uint32		delay = DELAY ; /* 1/25 * 1000 */

    SDL_Surface		*screen;
    SDL_Surface		*show ;
    SDL_Rect		dest;
    SDL_Event		event;

    /* handling ftp upload icon */
    SDL_Surface *ftp_upload_icon;
    SDL_Rect ftp_upload_dest;

    /* For debug */
    int nb_frame = 0 ; 
    time_t deb , end;
    /* For debug */

    /* FTP time info stat */
    time_t ftp_start, ftp_end;
    int ftp_time_state=0;
    unsigned long ftp_count=0;

    struct v4l2_cropcap 	cropcap;
    struct v4l2_crop 		crop;
    struct v4l2_format		format;

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif
    
    done = 0 ;

    
    /*
    if (argc != 5) {
        fprintf(stdout, "Usage: %s <input.bayer> <width> <height> <downscalefact>\n"
		"Where downscalefact must be 1,2 or 4 and width*downscalefact "
		"and height*downscalefact\nmust be multiple of 16 and <= than max res\n", 
		argv[0]);
		return 1;
    } else {
		fprintf( stdout , "sn9c102-webcam - v%s\n", VERSION );
		fprintf( stdout , "--------------------------------------\n" );
	}
    */

    
    fprintf( stdout , STR_DIV );
    fprintf( stdout , "sn9c102-webcam - v%s\n", VERSION );
    fprintf( stdout , STR_DIV );

    
    /* default values
    x=352;
    y=288;
    f=24;
    */

    /* reading configuration file */
    /* thanks to Christophe Lucas for his patch and help (lookee) */
    config=load_conf_from_file(CONF_FILE);
   

    x=config->x;
    y=config->y;
    f=config->bpp;
    
    /* check x,y (Igor Lopes Guerreiro workaround)*/
    if (!x||!y||!f)
    	fprintf(stdout, "xyf workaround (xyf default set)"); 
    if (!x)
    	x=352;
    if (!y)
    	y=288;
    if (!f)
    	f=24;
    /* end workaround */
    
    /* local file jpg */
    strncpy(config->local_file,"./cam.jpg",sizeof(config->local_file));

    /* build name of remote_tmp_file */
    /* trunk this name with NULL is bad... I'll improve the check of length of string */
    strncpy(config->remote_tmp_file,config->remote_file,sizeof(config->remote_tmp_file));
    strncat(config->remote_tmp_file,".tmp",sizeof(config->remote_tmp_file));
   

    /* check ftp_rate value */
    if (config->ftp_rate<1){
    	config->ftp_rate=FTP_RATE;
	printf("Incorrect ftp_rate value fixed to default %d\n",FTP_RATE);
    }
  
    /* check if ftp upload is enabled */
    ftp_upload=!strncmp(config->enable_ftp,"y",sizeof(config->enable_ftp));
    ftp_monitor=!strncmp(config->enable_ftp_monitor,"y", sizeof(config->enable_ftp_monitor));
    
    /* end copy import keys from file */
    
    /*
    x = atoi(argv[2]);
    y = atoi(argv[3]);
    f = atoi(argv[4]);
    */

    if (SDL_Init (SDL_INIT_VIDEO) < 0) {
		fprintf (stderr, "I can't init SDL\n");
		exit (1);
    }
    

    screen = SDL_SetVideoMode( x, y, 24, SDL_SWSURFACE|SDL_DOUBLEBUF|SDL_ANYFORMAT);
    if( screen == NULL ) {
		fprintf (stderr, "resolution error:  %ldx%ld and %d bpp \n", x , y , 24 );
		exit (1);
    }

    SDL_WM_SetCaption( "sn9c102-webcam", NULL );
	
    s = (char*)malloc(x * y);
    if( s == NULL ) {
		perror("Lack of memory !! :-(");
		exit(1);
    }

    d = (char*)malloc(x * y *3);
    if( d == NULL ) {
		perror("Lack of memory !! :-(");
		exit(1);
    }

    /* init ftp_upload_icon */
    ftp_upload_icon = SDL_LoadBMP(IMG_FTP_MONITOR);
    if ( ftp_upload_icon == NULL ) {
        fprintf(stderr, "ftp icon trouble: %s: %s\n",IMG_FTP_MONITOR, SDL_GetError());
	exit(1);
    }

    /* init dest */
    ftp_upload_dest.x = 352-ftp_upload_icon->w;
    ftp_upload_dest.y = 288-ftp_upload_icon->h;;
    ftp_upload_dest.w = ftp_upload_icon->w;
    ftp_upload_dest.h = ftp_upload_icon->h;

    fd1 = open(config->dev, O_RDONLY);
    if( fd1 == -1 ) {
		perror("Error : file does not exist (the module of webcam is loaded?)");
		exit(1);
    }

	/* only for debug (1) */
	/* printf("main: %s\n",config->host); */
	printf("We enumerate controls on this device:\nPlease wait...\n");
	enumerate_controls_on_device( fd1 ) ;
	
	/* only for debug (2) */
	/* printf("main: %s\n",config->host); */

	
    	fprintf( stdout , STR_DIV );
	fprintf( stdout, "res: %ld x %ld x %ld bpp \n",x,y,f);
	fprintf( stdout, "dev: %s \n",config->dev);

	/* print information about ftp */
	if (ftp_upload) {
		fprintf( stdout, "ftp upload rate: %d\n",config->ftp_rate);
		fprintf( stdout, "ftp host : %s \n", config->host);
		fprintf( stdout, "ftp file : %s \n", config->remote_file);
		fprintf( stdout, "ftp monitor : %s \n", ftp_upload ? "on" : "off");
	}
	
	
    	fprintf( stdout , STR_DIV );
	fprintf( stdout, "Now we play...\n" );
	fprintf( stdout , "\nKeypad: -/+ : gain\n" );
	fprintf( stdout , "u/j : red balance\n" );
	fprintf( stdout , "i/k : blue balance\n" );
	fprintf( stdout , "o/l : gamma balance\n" );
	fprintf( stdout , "p/m : brighness balance\n" );
	fprintf( stdout , "\nPress ESC to stop playing...\n\n" );



    cropcap.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if( ioctl(fd1,VIDIOC_CROPCAP,&cropcap) == -1 ) {
		fprintf(stderr, "VIDIOC_CROPCAP failed\n");
		perror("ioctl");
		exit(-1);
    }

    crop.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c.top = 0;
    crop.c.left = 0;
    crop.c.width = x*f;
    crop.c.height = y*f;
    if( ioctl(fd1,VIDIOC_S_CROP,&crop) == -1 ) {
    	fprintf(stderr, "VIDIOC_S_CROP failed\n");
		perror("ioctl");
		exit(-1);
    }

    format.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    format.fmt.pix.width = x;
    format.fmt.pix.height = y;
    format.fmt.pix.pixelformat = V4L2_PIX_FMT_SBGGR8;
    if( ioctl(fd1,VIDIOC_S_FMT,&format) == -1 ) {
    	fprintf(stderr, "VIDIOC_S_FMT failed\n");
		perror("ioctl");
		exit(-1);
    }

    time( &deb );

    while( !done ) {
		int ret ;

		ret = read(fd1, s, x * y);
		if( ret != (x * y ) ) {
		    printf("We will have fragmented picture :-(\n" );
		}
		nb_frame ++ ;
		bayer2rgb24(d, s, x, y);
		show = SDL_CreateRGBSurfaceFrom( (void *)d, x, y, 24, x*3, rmask, gmask, bmask, 0 );
		if( show == NULL ) {
		    printf("Creation of SLD_Surface is failed\n");
		} else {
		    dest.x = 0 ; 
		    dest.y = 0 ;
		    dest.w = x ;
		    dest.h = y ;
   	 
		    if( SDL_BlitSurface( show, NULL, screen, &dest) != 0 ) {
				printf("SDL_BlitSurface failed.\n");
		    }

		   
		    SDL_Flip( screen );
		    
		    
		    
			/* controllo sul rate di upload o salvataggio */
			if(ftp_rate_counter>=config->ftp_rate)
				ftp_rate_counter=1;
			
			/* save image */
			if ((save_img|ftp_upload)&&(ftp_rate_counter++==1)){

				/* save a surface on BMP format to file (cam.bmp) */
				/* SDL_SaveBMP(screen, "cam.bmp"); */
				interface_capture(screen, 352, 288);

				
				/* convert BMP to JPG format (cam.jpg) */
				/* passo argv perche imagemagick lo vuole  (?!?!?!) */
				/* bmp2jpg(argv); */
		

				/* upload FTP dell'immagine (se richiesto) */
				if (ftp_upload&&ftp_rate_counter){
					
					if (ftp_time_state==0)
						time(&ftp_start);
			
					
					/* show only if ftp_monitor is enabled */
					if (ftp_monitor){
						/* show ftp upload */
			   			SDL_UpdateRects(screen, 1, &ftp_upload_dest);
         	    				SDL_BlitSurface(ftp_upload_icon, NULL, screen, &ftp_upload_dest);
						SDL_Flip(screen);
		    				/* end show ftp upload */

					}
				
					
					/* upload file */
					ftpup(config);
					ftp_count++;

				
					if(ftp_time_state==1){
						time(&ftp_end);
						printf("image FTP upload rate: %lg sec.\n" , difftime( ftp_end , ftp_start) );
						/* end ftp stat */
						ftp_time_state=2;
					}

					if(ftp_time_state==0)
						ftp_time_state++;
				}
			}
		}
		
		while ( SDL_PollEvent(&event) ) {
		    switch (event.type) {
			case SDL_KEYDOWN:
			    if ( event.key.keysym.sym == SDLK_ESCAPE ) {
					done = 1;
			    } else if( event.key.keysym.sym == SDLK_KP_MINUS ) {
					change_ctrl( fd1, V4L2_CID_GAIN, MINUS );
				} else if( event.key.keysym.sym == SDLK_KP_PLUS ) {
					change_ctrl( fd1 , V4L2_CID_GAIN, PLUS );
				} else if( event.key.keysym.sym == SDLK_p ) {
					change_ctrl( fd1 , V4L2_CID_BRIGHTNESS, PLUS );
				} else if( event.key.keysym.sym == SDLK_m ) {
					change_ctrl( fd1 , V4L2_CID_BRIGHTNESS, MINUS );
				} else if( event.key.keysym.sym == SDLK_u ) {
					change_ctrl( fd1 , V4L2_CID_RED_BALANCE, PLUS );
				} else if( event.key.keysym.sym == SDLK_j ) {
					change_ctrl( fd1 , V4L2_CID_RED_BALANCE, MINUS );
				} else if( event.key.keysym.sym == SDLK_i ) {
					change_ctrl( fd1 , V4L2_CID_BLUE_BALANCE, PLUS );
				} else if( event.key.keysym.sym == SDLK_k ) {
					change_ctrl( fd1 , V4L2_CID_BLUE_BALANCE, MINUS );
				} else if( event.key.keysym.sym == SDLK_o ) {
					change_ctrl( fd1 , V4L2_CID_GAMMA, PLUS );
				} else if( event.key.keysym.sym == SDLK_l ) {
					change_ctrl( fd1 , V4L2_CID_GAMMA, MINUS );
				}
			    break;
			case SDL_QUIT:
			    done = 1;
			break;
			default:
			break;
		    }
		}   
		SDL_Delay( delay ) ;
    }
    time( &end );
    printf("I have been playing %d frame(s) in %lg second(s)\n" , nb_frame, difftime( end , deb ) );
    printf("I uploaded %ld image(s) by FTP\n",ftp_count);
    close(fd1);
    SDL_FreeSurface( screen );
    SDL_FreeSurface( show ) ;
    SDL_Quit();
return 0;
}
